﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NLog;
using System.Xml.Serialization;
using System.IO;
using System.Net.Sockets;

namespace ChatSystem
{
    /// <summary>
    /// swb - modified!!!
    /// </summary>
    public class MessageBuffer
    {
        #region Constants

        public enum BufferType : byte
        {
            INVALID = 0,
            USERNAME, // 1 Field: User
            DATA,      // 2 Fields: User, Contents
            DISCONNECT  // No Fields, existing data used
        };

        public const byte FieldSeparator = 0;  // Separate the fields in DATA

        public static string CLASS_NAME = typeof(MessageBuffer).ToString();

        // Effectively a constant
        private static Logger logger = LogManager.GetCurrentClassLogger();

        #endregion

        #region Public Access

        #region Public Properties

        public BufferType MessageTypeValue
        {
            get { return mMessageType; }
            set { mMessageType = value; }
        }

        public string ContentsValue
        {
            get { return mContents; }
            set { mContents = value; }
        }

        public string UserValue
        {
            get { return mUser; }
            set { mUser = value; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Set default values for all fields.
        /// </summary>
        public MessageBuffer()
        {
            Reset();
        }

        /// <summary>
        /// Set all fields according to arguments
        /// </summary>
        /// <param name=""></param>
        public MessageBuffer(BufferType aMessageType, string aUser, string aContents)
        {
            Set(aMessageType, aUser, aContents);
        }

        /// <summary>
        /// Copy Constructor
        /// </summary>
        /// <param name=""></param>
        public MessageBuffer(MessageBuffer aItem)
        {
            Set(aItem);
        }

        #endregion

        #region Reset, Set

        /// <summary>
        /// Set fields to default values
        /// </summary>
        public void Reset()
        {
            mMessageType = BufferType.INVALID;
            mUser = "";
            mContents = "";
        }

        /// <summary>
        /// Set the values of every field according to arguments
        /// </summary>
        /// <param name=""></param>
        public void Set(BufferType aMessageType, string aUser, string aContents)
        {
            mMessageType = aMessageType;
            mUser = aUser;
            mContents = aContents;
        }

        /// <summary>
        /// Copy our values from given MessageBufferItem
        /// </summary>
        /// <param name="aFrom"></param>
        public void Set(MessageBuffer aFrom)
        {
            mMessageType = aFrom.mMessageType;
            mUser = aFrom.mUser;
            mContents = aFrom.mContents;
        }

        #endregion

        #region Logging

        /// <summary>
        /// Return ToString for this class, or "null"
        /// </summary>
        /// <param name="aValue"></param>
        /// <returns></returns>
        public static string ToStringP(MessageBuffer aValue)
        {
            if (aValue == null)
                return "null";
            else
                return aValue.ToString();
        }

        /// <summary>
        /// Return fields of this object as a string
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "mMessageType=" + mMessageType.ToString() + ", mUser='" + mUser + "', mContents='" + mContents + "'";
        }

        /// <summary>
        /// Log this object to nlog
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public void Logit(LogLevel aLevel, string aIndent, string aHeader)
        {
            string logString = LogitToString(aLevel, aIndent, aHeader);
            logger.Log(aLevel, logString);
        }

        /// <summary>
        /// Log this object to nlog - short version
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public void LogitL(LogLevel aLevel, string aIndent, string aHeader)
        {
            string logString = LogitLToString(aLevel, aIndent, aHeader);
            logger.Log(aLevel, logString);
        }

        /// <summary>
        /// Log the contents of this class to String
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public string LogitToString(LogLevel aLevel, string aIndent, string aHeader)
        {
            // Don't log if not within level
            if (IdLibrary.WithinLoggerLevel(aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();

            sb.AppendLine(aIndent + aHeader);
            sb.AppendLine(aIndent + "    MessageBuffer:");
            sb.AppendLine(aIndent + "        " + "MessageType=" + mMessageType + ", User='" + mUser + "'");
            sb.AppendLine(aIndent + "        " + "Contents='" + mContents + "'");

            sb.Append(aIndent + "    End of MessageBuffer");
            return sb.ToString();
        }

        /// <summary>
        /// Log the contents of this class to String - short version
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public string LogitLToString(LogLevel aLevel, string aIndent, string aHeader)
        {
            // Don't log if not within level
            if (IdLibrary.WithinLoggerLevel(aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();
            string result = "Type=" + mMessageType + ", User='" + mUser + "', Msg='" + GeneralRoutines.ShortString(mContents, 30) + "'";

            if (aHeader.Length > 0)
                sb.AppendLine(aIndent + aHeader + " " + result);
            else
                sb.AppendLine(aIndent + result);

            return sb.ToString();
        }

        #endregion

        #region I/O Routines

        /// <summary>
        /// Convert our class to an XML string via Serialize. Can throw exceptions
        /// </summary>
        /// <returns></returns>
        public string ToXML()
        {
            try
            {
                // Force UTF-8
                XmlSerializer serializer = new XmlSerializer(typeof(MessageBuffer));
                using (StringWriter writer = new IdLibrary.Utf8StringWriter())
                {
                    serializer.Serialize(writer, this);
                    return writer.ToString();
                }
            }
            catch (Exception ex)
            {
                logger.Error("ToXML: Exception '" + ex + "'");
                throw new Exception("ToXML", ex);
            }
        }

        /// <summary>
        /// Convert given xml string to our class. Throw exception on fail.
        /// </summary>
        /// <param name="xml"></param>
        /// <returns></returns>
        public MessageBuffer FromXML(string xml)
        {
            MessageBuffer ei = null;
            if ((xml != null) && (xml.Length > 0))
            {
                try
                {
                    XmlSerializer serializer = new XmlSerializer(typeof(MessageBuffer));
                    ei = (MessageBuffer)serializer.Deserialize(new StringReader(xml));
                }
                catch (Exception ex)
                {
                    logger.Error("FromXML: Got exception '" + ex + "'");
                    throw new Exception("FromXML", ex);
                }
            }
            return ei;
        }

        /// <summary>
        /// Write out our class to given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool WriteFile(string fileName)
        {
            try
            {
                StreamWriter writer = new StreamWriter(fileName);
                writer.Write(this.ToXML());
                writer.Close();
            }
            catch (Exception ex)
            {
                logger.Info("WriteFile: Got exception '" + ex + "'");
                return false;
            }
            // Success
            return true;
        }

        /// <summary>
        /// Read our class from given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool ReadFile(string fileName)
        {
            try
            {
                StreamReader reader = new StreamReader(fileName);
                string contents = reader.ReadToEnd();
                reader.Close();

                // Set all our values from what we got back
                MessageBuffer fi = FromXML(contents);
                this.Set(fi);
            }
            catch (Exception ex)
            {
                logger.Info("ReadFile: Got exception '" + ex + "'");
                return false;
            }

            // Success
            return true;
        }

        #endregion

        #region Tester

        /// <summary>
        /// Basic test routine for MessageBufferItem.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool Tester()
        {
            string TEST_FILE = "testMessageBuffer.xml";
            bool result;
            MessageBuffer fi1, fi2, fi3;

            logger.Info("");
            logger.Info("Tester:1:");

            // Test constructors
            try
            {
                fi1 = new MessageBuffer();
                fi2 = BuildSample(BufferType.DATA);
                fi3 = new MessageBuffer(fi2);
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Constructor test got ex: " + ex);
                return false;
            }

            // Test logging
            try
            {
                logger.Info("Tester: ToStringP(fi2)=[" + ToStringP(fi2) + "]");
                logger.Info("");
                fi2.Logit(LogLevel.Info, "", "Tester: Logit without indent:");
                fi3.LogitL(LogLevel.Info, "   ", "Tester: LogitL with indent:");
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Logit test got ex: " + ex);
                return false;
            }

            // Test Write and read
            try
            {
                logger.Info("");

                // Write out to XML
                if (fi2.WriteFile(TEST_FILE) == false)
                {
                    logger.Info("Tester: WriteFile(" + TEST_FILE + ") failed, returning false");
                    return false;
                }

                // Read in as different object
                result = fi1.ReadFile(TEST_FILE);

                if (result != true)
                {
                    logger.Info("Tester: Read " + TEST_FILE + " failed, returning false");
                    return false;
                }

                fi1.Logit(LogLevel.Info, "", "Tester: Read " + TEST_FILE + ":");
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Read/Write test got ex: " + ex);
                return false;
            }

            // Success
            logger.Info("Tester: tests completed successfully");
            return true;
        }

        /// <summary>
        /// Return a sample record MessageBufferItem.
        /// Return sample on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static MessageBuffer BuildSample(BufferType aType)
        {
            logger.Info("BuildSample: aType='" + aType + "'");
            BufferType tempMessageType;
            string tempContents;

            MessageBuffer rec;
            string lowerID = Guid.NewGuid().ToString();

            if (aType != BufferType.INVALID)
            {
                tempMessageType = aType;
                tempContents = "aaa";

                rec = new MessageBuffer(tempMessageType, "user", tempContents);
            }
            else
            {
                rec = new MessageBuffer(BufferType.INVALID, "user", "unused");
            }

            return rec;
        }

        /// <summary>
        /// Drill down through sample record until non-class 2nd item and update it
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public bool UpdateSample()
        {
            logger.Info("UpdateSample:1: mMessageType = USERNAME");
            mMessageType = BufferType.USERNAME;
            return true;
        }

        #endregion

        #region Class-Specific

        /// <summary>
        /// Load our contents from byte array.
        /// Return true on success, false on fail.
        /// </summary>
        /// <param name="aNs"></param>
        /// <returns></returns>
        public bool LoadFromByteArray(byte[] aInput)
        {
            BufferType msgType;
            string userField, contentsField;

            // Check obvious problem
            if (aInput == null || aInput.Length < 1)
            {
                logger.Error("LoadFromByteArray: Got null or empty input");
                return false;
            }

            // Get the Message type
            msgType = ByteToBufferType(aInput[0]);
            if (msgType == BufferType.INVALID)
            {
                logger.Error("LoadFromByteArray: Got INVALID message type");
                return false;
            }

            // Extract rest of the message
            if (SplitByFields(aInput, msgType, out userField, out contentsField) == false)
            {
                logger.Error("LoadFromByteArray: Unable to split by fields");
                return false;
            }

            // fill out record
            Set(msgType, userField, contentsField);
            return true;
        }

        /// <summary>
        /// Split our message into its component fields by MessageType
        /// </summary>
        /// <param name="aInput"></param>
        /// <param name="aType"></param>
        /// <param name="aUserField"></param>
        /// <param name="aContentsField"></param>
        /// <returns></returns>
        private bool SplitByFields(byte[] aInput, MessageBuffer.BufferType aType, out string aUserField, out string aContentsField)
        {
            List<byte[]> fields;
            byte[] field;
            int i, curIdx = 1;

            // Initialize, needs to be before try
            aUserField = "UserError";
            aContentsField = "ContentsError";

            try
            {
                fields = new List<byte[]>();

                // Each section is terminated by a zero. These are ascii strings, so they should not have a 0
                for (i = curIdx; i < aInput.Length; i++)
                {
                    if (aInput[i] == 0) // Start of new field
                    {
                        field = new byte[i - curIdx];
                        System.Buffer.BlockCopy(aInput, curIdx, field, 0, field.Length);
                        fields.Add(field);
                        curIdx = i + 1;
                    }
                }

                if (curIdx <= i)
                {
                    field = new byte[i - curIdx];
                    System.Buffer.BlockCopy(aInput, curIdx, field, 0, field.Length);
                    fields.Add(field);
                }

                // fields is a list of byte arrays, each one represents a field to be converted back to a string
                // [0]=User, [1] = Contents
                if (fields.Count < 2)
                {
                    logger.Error("SplitByFields: Record has less than 2 fields, returning false");
                    return false;
                }

                aUserField = ASCIIEncoding.ASCII.GetString(fields[0]);
                aContentsField = ASCIIEncoding.ASCII.GetString(fields[1]);

                logger.Info("SplitByFields:3: returning aUserField='" + aUserField + "', aContentsField='"+aContentsField+"'");
                return true;
            }
            catch (Exception ex)
            {
                logger.Error("SplitByFields: got ex: " + ex);
                return false;
            }
        }

        /// <summary>
        /// Return byte array corresponding to this message
        /// </summary>
        /// <returns></returns>
        public byte[] SaveToByteArray()
        {
            byte[] returnValue;
            byte[] user, contents;

            // Get length of content part
            user = Encoding.ASCII.GetBytes(mUser);
            contents = Encoding.ASCII.GetBytes(mContents);

            // Create appropriately-sized return buffer, and load the message type into it
            // Type User Zero Contents
            int bufLength = user.Length + contents.Length + 2;
            returnValue = new byte[bufLength];
            logger.Info("SaveToByteArray: uLen="+user.Length+", cLen="+contents.Length+", bufLen="+ bufLength);
            returnValue[0] = (byte)mMessageType;

            // Add the rest of our byte array
            System.Buffer.BlockCopy(user, 0, returnValue, 1, user.Length);
            returnValue[user.Length+1] = 0;  // field separator

            int newPos = user.Length + 2;
            System.Buffer.BlockCopy(contents, 0, returnValue, newPos, contents.Length);

            LogByteArray(returnValue, LogLevel.Info, "", "SaveTo: Bytes to send:");
            return returnValue;
        }

        public static void LogByteArray() { }

        /// <summary>
        /// Log byte string to nlog
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public static void LogByteArray(byte[] aValue, LogLevel aLevel, string aIndent, string aHeader)
        {
            string logString = LogByteArrayToString(aValue, aLevel, aIndent, aHeader);
            logger.Log(aLevel, logString);
        }

        /// <summary>
        /// Log the contents of this class to String
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public static string LogByteArrayToString(byte[] aValue, LogLevel aLevel, string aIndent, string aHeader)
        {
            int bytesPerLine = 16;

            // Don't log if not within level
            if (IdLibrary.WithinLoggerLevel(aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();
            int i;
            StringBuilder explainSection = new StringBuilder(), valueSection = new StringBuilder();

            sb.AppendLine(aIndent + aHeader + "    Byte Count=" + aValue.Length);

            for (i = 0; i < aValue.Length; i++)
            {
                if (i % bytesPerLine == 0)
                {
                    if (i != 0)
                        sb.AppendLine(String.Format(" {0}    {1}", valueSection.ToString(), explainSection.ToString()));

                    explainSection.Clear();
                    valueSection.Clear();
                    string indexHeading = String.Format("[{0,7}]", i);
                    sb.Append(indexHeading);
                }

                //decValue = aValue[i].ToString("D3");

                // Old fashioned way, assumes ascii
                if (aValue[i] >= 32 && aValue[i] < 127)
                {
                    explainSection.Append(((char)aValue[i]).ToString());
                    //explain += " or "+ ASCIIEncoding.ASCII.GetString(aValue, i, 1);
                }
                else
                    explainSection.Append(".");

                valueSection.Append(String.Format(" {0,3}", aValue[i]));

            }

            // Finish last line
            if (valueSection.Length > 0)
            {
                int bytesLeft = bytesPerLine - (i % bytesPerLine);
                valueSection.Append("                                                                                               ".Substring(0,
                    4 * bytesLeft));
                sb.AppendLine(String.Format(" {0}    {1}", valueSection.ToString(), explainSection.ToString()));
            }
            else
                sb.AppendLine();

            return sb.ToString();
        }


        /// <summary>
        /// Return buffer type corresponding to this byte, or INVALID
        /// </summary>
        /// <param name="aType"></param>
        /// <returns></returns>
        public static BufferType ByteToBufferType(byte aType)
        {
            BufferType value;

            try
            {
                value = (BufferType)aType;
            }
            catch (Exception)
            {
                value = BufferType.INVALID;
            }

            return value;
        }

        #endregion

        #endregion

        #region Private Access

        #region Private Members

        private BufferType mMessageType;
        private string mUser;   // User message originally from
        private string mContents;

        #endregion

        #endregion

    }
}
