﻿using System;
using System.Windows.Forms;
using System.Text;
using System.Net.Sockets ;
using System.Threading;
using NLog;

namespace ChatSystem
{
    public partial class Form1 : Form
    {
        #region Constants

        // Effectively a constant
        private static Logger logger = LogManager.GetCurrentClassLogger();

        #endregion

        #region Member Fields

        TcpClient clientSocket = new System.Net.Sockets.TcpClient();
        NetworkStream serverStream = null; // default(NetworkStream);

        string msgValue = "";
        Thread ctThread = null;
        string myID;

        #endregion

        #region Constructors

        public Form1()
        {
            InitializeComponent();

            myID = Environment.MachineName + "-user";
            logger.Info("Form1: Initial myID='" + myID + "'");
        }

        #endregion

        #region getMessage

        private void getMessage()
        {
            int available;

            try
            {
                while (true)
                {
                    try
                    {
                        available = clientSocket.Available;
                    }
                    catch (ObjectDisposedException)
                    {
                        // Socket connection dropped
                        DisplayUpdate("Connection dropped");
                        return;
                    }
                    catch (Exception ex)
                    {
                        DisplayUpdate("Socket access got ex: "+ex);
                        return;
                    }

                    if (available > 0)
                    {
                        // Read from stream
                        byte[] inBytes = new byte[available];
                        serverStream = clientSocket.GetStream();

                        serverStream.Read(inBytes, 0, available);
                        MessageBuffer.LogByteArray(inBytes, LogLevel.Info, "", "getMessage: Byte read:");

                        // Convert to message
                        MessageBuffer buf = new MessageBuffer();
                        buf.LoadFromByteArray(inBytes);
                        buf.LogitL(LogLevel.Info, "", "getMessage: Buffer:");

                        //string returndata = System.Text.Encoding.ASCII.GetString(inBytes);
                        //logger.Info("getMessage: returndata='" + returndata + "'");

                        DisplayUpdate(buf);  // in case returndata is null
                    }
                    else
                        Thread.Sleep(100);
                }
            }
            catch (Exception ex)
            {
                DisplayUpdate("getMessage: Got ex: " + ex);
            }
        }

        #endregion

        #region Support

        /// <summary>
        /// Display a system message in center area
        /// </summary>
        /// <param name="aMessage"></param>
        private void DisplayUpdate(string aMessage)
        {
            MessageBuffer buf = new MessageBuffer(MessageBuffer.BufferType.DATA, "Client", aMessage);
            DisplayUpdate(buf);
        }

        /// <summary>
        /// Display the given string in the center area.
        /// </summary>
        /// <param name="aMessage"></param>
        private void DisplayUpdate(MessageBuffer aMessage)
        {
            string header, body;

            // Fill out header
            if (aMessage.UserValue.Length > 0)
                header = aMessage.UserValue + ">>";
            else
                header = ">>";

            // Fill out body
            if (aMessage.ContentsValue.Length > 0)
                body = aMessage.ContentsValue;
            else
                body = "";
            
            // msg() can't take arguments, do the variable way
            msgValue = header + "   " + body;
            logger.Info("DisplayUpdate: msgValue='" + msgValue + "'");

            // Put to screen, via Invoking if needed
            msg();
        }

        /// <summary>
        /// Add to the center textbox, calling Invoker if needed.
        /// </summary>
        private void msg()
        {
            try
            {
                if (this.InvokeRequired)
                    this.Invoke(new MethodInvoker(msg));
                else
                    textBox2.Text = textBox2.Text + Environment.NewLine + msgValue;
            }
            catch (ObjectDisposedException)
            {
                // We are shutting down anyway
            }
        }

        /// <summary>
        /// Format and send the output to the server
        /// </summary>
        /// <param name="aMsg"></param>
        /// <returns></returns>
        private bool WriteMessageToServer(ChatSystem.MessageBuffer aMsg)
        {
            if (serverStream == null)
            {
                DisplayUpdate("*** Can't write to server ***");
                return false;
            }

            aMsg.LogitL(LogLevel.Info, "", "WriteMessageToServer:1:");
            byte[] outBytes = aMsg.SaveToByteArray();

            //byte[] outStream = System.Text.Encoding.ASCII.GetBytes(aMsg + "$");
            serverStream.Write(outBytes, 0, outBytes.Length);
            serverStream.Flush();

            return true;
        }

        #endregion

        #region Events

        /// <summary>
        ///  Write to server button
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void SendButton_Click(object sender, EventArgs e)
        {
            logger.Info("SendButton_Click:1: textBox3='" + textBox3.Text+"'");

            // User name will actually be filled out by server
            MessageBuffer mb = new MessageBuffer(MessageBuffer.BufferType.DATA, "unused", textBox3.Text);
            WriteMessageToServer(mb);

            textBox3.Text = "";
        }

        /// <summary>
        /// Connect button
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ConnectButton_Click(object sender, EventArgs e)
        {
            try
            {
                logger.Info("ConnectButton_Click:1: server='" + serverIpBox.Text + "', port="+GeneralRoutines.DEFAULT_SERVER_PORT);
                DisplayUpdate("Connecting to Chat Server "+ serverIpBox.Text+" ...");

                clientSocket.Connect(serverIpBox.Text, GeneralRoutines.DEFAULT_SERVER_PORT);
                serverStream = clientSocket.GetStream();

                myID = Environment.MachineName + "-"+textBox1.Text;
                logger.Info("ConnectButton_Click:2: myID='" + myID + "'");

                MessageBuffer mb = new MessageBuffer(MessageBuffer.BufferType.USERNAME, myID, "unused");
                WriteMessageToServer(mb);

                ctThread = new Thread(getMessage);
                ctThread.Start();
            }
            catch (Exception ex)
            {
                serverStream = null;
                DisplayUpdate("Unable to connect to server, msg=" + ex.Message);
            }
        }

        /// <summary>
        /// Do any shutdown actions
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            try
            {
                logger.Info("Form1_FormClosing:1:");

                // Try to warn server
                if (serverStream != null)
                {
                    MessageBuffer mb = new MessageBuffer(MessageBuffer.BufferType.DISCONNECT, "unused", "ExitExitExit");
                    WriteMessageToServer(mb);
                    serverStream.Flush();
                    serverStream.Close();
                }

                // Shut down message receive thread
                logger.Info("Form1_FormClosing:2: ctThread=" + (ctThread == null ? "null" : "set"));

                if (ctThread != null)
                {
                    logger.Info("Form1_FormClosing:3: Doing Abort");
                    ctThread.Abort();
                    logger.Info("Form1_FormClosing:4: Finished Abort");
                    ctThread = null;
                }
            }
            catch (Exception ex)
            {
                logger.Error("Form1_FormClosing: got ex: " + ex);
            }
        }

        #endregion
    }
}

