﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net.Sockets;
using System.Net;
using System.Threading;
using NLog;

namespace ChatSystem
{
    class ScottChatServer
    {
        #region Constants

        // Put listening in a separate thread
        delegate void mThread(ref bool isRunning);
        delegate void AccptTcpClnt(ref TcpClient client, TcpListener listener);

        // Effectively a constant
        private static Logger logger = LogManager.GetCurrentClassLogger();

        #endregion

        ListenWorker workerObject;

        /// <summary>
        ///  Default constructor
        /// </summary>
        public ScottChatServer()
        {
            logger.Info("ScottChatServer:1:");
            Utility.mChatUsers = new ChatUserList();
            workerObject = null;
        }

        /// <summary>
        /// Entry point from Main()
        /// </summary>
        public void Run()
        {
            string initialLog = String.Format("I'm {0} at {1}...", IPAddress.Any, GeneralRoutines.DEFAULT_SERVER_PORT);
            logger.Info("Run:1: " + initialLog);
            Console.WriteLine(initialLog);

            // Start our TCP listener in the background
            StartBackgroundListener();

            // For life of program, read the sockets and broadcast results
            ReadLoop();

            // Shut down the listener, by force if necessary
            workerObject.RequestStop();
            Thread.Sleep(1500);

            if (workerObject.mIsAlive == true)
                workerObject.Abort();

            Console.WriteLine("Run: Finished");
            Console.ReadKey();
        }

        /// <summary>
        /// Start Accepting TCP requests on a worker thread
        /// </summary>
        private void StartBackgroundListener()
        {
            logger.Info("StartBackgroundListener:1:");

            // Create the thread object. This does not start the thread.
            ListenWorker workerObject = new ListenWorker();
            Thread workerThread = new Thread(workerObject.DoWork);

            // Start the worker thread.
            workerThread.Start();
            Console.WriteLine("Run: Started worker thread...");

            // Give it time to start
            Thread.Sleep(1500);

            logger.Info("StartBackgroundListener: returning success");
        }

        /// <summary>
        /// Main loop, read from the TcpClients and broadcast their messages
        /// </summary>
        private void ReadLoop()
        {
            bool foundAtLeastOne;

            while (true)
            {
                foundAtLeastOne = false;

                foreach (ChatUserItem user in Utility.mChatUsers)
                {
                    if (user.IsValidValue == false)  // In process of being deleted
                        continue;

                    if (user.ClientValue.Available > 0)
                    {
                        foundAtLeastOne = true;
                        HandleInput(user);

                    }
                }

                // Cleanup
                if (Utility.mChatUsers.GetInvalidIndex() >= 0)
                    lock (Utility.mChatUsers)
                    {
                        Utility.mChatUsers.DeleteInvalidItems();
                    }

                // Keep from spinning excessively
                if (foundAtLeastOne == false)
                    Thread.Sleep(100);
            }
        }

        /// <summary>
        /// Read in message from this user's socket and act on it.
        /// </summary>
        /// <param name="user"></param>
        private void HandleInput(ChatUserItem user)
        {
            MessageBuffer mb;

            logger.Info("HandleInput:1: user=" + user.IdTagValue);

            // Read in the message
            if ((mb = user.ReadStream()) == null)
            {
                Utility.ConsoleLog(LogLevel.Error, "HandleInput: Read from '" + user.IdTagValue + "' failed, user will be deleted");
                return;
            }

            // Do actual processing according to message type
            switch (mb.MessageTypeValue)
            {
                case MessageBuffer.BufferType.DATA:
                    BroadcastData(user.IdTagValue, mb);
                    break;

                case MessageBuffer.BufferType.USERNAME:
                    Utility.ConsoleLog(LogLevel.Info, "Setting ID '" + user.IdTagValue + "' to user '" + mb.UserValue + "'");
                    user.IdTagValue = mb.UserValue;

                    mb.Set(MessageBuffer.BufferType.DATA, "Server", "You are connected.");
                    if (user.WriteStream(mb) == false)
                    {
                        Utility.ConsoleLog(LogLevel.Error, "HandleInput: Cannot write to " + user.IdTagValue);
                    }
                    break;

                case MessageBuffer.BufferType.DISCONNECT:
                    MessageBuffer mbDisconnect = new MessageBuffer(MessageBuffer.BufferType.DISCONNECT, "Server", user.IdTagValue + " has disconnected");
                    BroadcastData("Server", mbDisconnect);
                    break;
            }

            logger.Info("HandleInput: returning");
        }

        /// <summary>
        /// Write message out to all users except one who sent
        /// </summary>
        /// <param name="aIdTag"></param>
        /// <param name="aBuf"></param>
        private void BroadcastData(string aIdTag, MessageBuffer aBuf)
        {
            logger.Info("BroadcastData:1: sender='" + aIdTag + "' sending msg '" + aBuf.ContentsValue + "'");

            // Need to add who it's from
            MessageBuffer sendMb = new MessageBuffer(aBuf.MessageTypeValue, aIdTag, aBuf.ContentsValue);

            foreach (ChatUserItem user in Utility.mChatUsers)
            {
                if (user.IsValidValue == false)
                    continue;

                // Send to them
                if (user.WriteStream(sendMb) == false)
                {
                    Utility.ConsoleLog(LogLevel.Error, "Cannot write to " + user.IdTagValue);
                }
            }
        }
    }
}
